// ============================================================================ //
// Lightflow Rendering Interface                                                //
// ============================================================================ //
// LfTransform.h                                                                //
// ============================================================================ //
//                                                                              //
// The contents of this file are subject to the Lightflow Rendering Interface   //
// License (the License); you may not use this file except in compliance with   //
// the License. You may obtain a copy of the License by sending an email at     //
// jp@lightflowtech.com with subject LRI-LICENSE.                               //
//                                                                              //
// ============================================================================ //
// Copyright (c) 1995-2000 by Marco and Jacopo Pantaleoni. All rights reserved. //
// ============================================================================ //

// N O T E
//
// These classes implement 3D transformations. 'LfTransform' class implements transformations
// for points and vectors, but not for object normals. In fact, these require special treatment
// (see for example "Advanced Animation and Rendering Techniques" by A. Watt, M. Watt, pg. 6/7)
// This treatment is implemented by 'LfNTransform' class.
// Transformations are implemented using 4x4 matrices, built on 4D vectors. This allows to
// access the elements with the sintax M[row][column].
// The convention followed is that transformations apply to column vectors, so matrices left
// multiply by vectors, and composition follows the rule:
//
// T1, T2 are transformations, so: T2(T1(v)) == (T2*T1)(v).

#ifndef _LfTransform_h
#define _LfTransform_h 1

#include <Lightflow/LfVector.h>
#include <iostream.h>

class LfTransform;		// 3D transformation
class LfNTransform;		// 3D transformation with transf. for normals

// In the following, these rules apply:
//
// P prefix:   Point
// V prefix:   Vector
// N prefix:   Normal
// NL suffix:  Non Linear transformation (with final divide)
//

class LfAPI_CS LfTransform
{
private:
	LfMatrix4 t;

public:
inline                      LfTransform			();
inline                      LfTransform			(const LfTransform&);
inline                      LfTransform			(const LfNTransform&);
inline                      LfTransform			(const LfMatrix4&);

inline		  LfTransform&  operator  =			(const LfTransform&);

			  LfTransform&  Zero				();
			  LfTransform&  Identity			();
			  LfTransform&  Scaling				(const LfVector&);
			  LfTransform&  Translation			(const LfVector&);
			  LfTransform&  AlignmentToZ		(const LfVector&);
			  LfTransform&  AlignmentTo			(const LfVector&, const LfVector&);
			  LfTransform&  RotationAroundX		(LfFloat);
			  LfTransform&  RotationAroundY		(LfFloat);
			  LfTransform&  RotationAroundZ		(LfFloat);
			  LfTransform&  RotationAroundAxis	(const LfPoint&, const LfVector&, LfFloat);
			  LfTransform&  ReflectionOnXY		();
			  LfTransform&  WorldToCamera		(const LfPoint& C, const LfVector& N, \
												 const LfVector& UpApp);
			  LfTransform&  Perspective			(LfFloat w, LfFloat h, LfFloat F, LfFloat B);

friend inline LfAPI_CS int           operator == 		(const LfTransform&, const LfTransform&);
friend inline LfAPI_CS int           operator != 		(const LfTransform&, const LfTransform&);
friend inline LfAPI_CS LfTransform&  Compose     		(const LfTransform&, const LfTransform&, \
												 LfTransform&);
friend inline LfAPI_CS LfTransform&  Invert      		(const LfTransform&, LfTransform&);
friend inline LfAPI_CS LfPoint&      PTrans				(const LfTransform&, const LfPoint&,  LfPoint&);
friend inline LfAPI_CS LfPoint&      PTransNL			(const LfTransform&, const LfPoint&,  LfPoint&);
friend inline LfAPI_CS LfVector&     VTrans				(const LfTransform&, const LfVector&, LfVector&);

friend LfAPI_CS ostream& operator<< (ostream& s, LfTransform t);
friend LfAPI_CS istream& operator>> (istream& s, LfTransform& t);

friend class LfNTransform;

inline		  LfVector4&	   operator [] (int); // Added on 14/7/97
inline        const LfVector4& Get(int) const;
inline        const LfMatrix4& GetMatrix(void) const;
};

class LfAPI_CS LfNTransform
{
private:
	LfMatrix4 t;
	LfMatrix3 n;

public:
inline                      LfNTransform			();
inline                      LfNTransform			(const LfNTransform&);
inline                      LfNTransform			(const LfTransform&);
inline                      LfNTransform			(const LfMatrix4&);

inline		  LfNTransform& operator  =			(const LfNTransform&);

			  LfNTransform&	Zero				();
			  LfNTransform&	Identity			();
			  LfNTransform&	Scaling				(const LfVector&);
			  LfNTransform&	Translation			(const LfVector&);
			  LfNTransform&	AlignmentToZ		(const LfVector&);
			  LfNTransform&	AlignmentTo			(const LfVector&, const LfVector&);
			  LfNTransform&	RotationAroundX		(LfFloat);
			  LfNTransform&	RotationAroundY		(LfFloat);
			  LfNTransform&	RotationAroundZ		(LfFloat);
			  LfNTransform&	RotationAroundAxis	(const LfPoint&, const LfVector&, LfFloat);
			  LfNTransform&	ReflectionOnXY		();
			  LfNTransform&	WorldToCamera		(const LfPoint& C, const LfVector& N, \
												 const LfVector& UpApp);
			  LfNTransform&	Perspective			(LfFloat w, LfFloat h, LfFloat F, LfFloat B);

friend inline LfAPI_CS int           operator == 		(const LfNTransform&, const LfNTransform&);
friend inline LfAPI_CS int           operator != 		(const LfNTransform&, const LfNTransform&);
friend inline LfAPI_CS LfNTransform& Compose				(const LfNTransform&, const LfNTransform&, \
												 LfNTransform&);
friend inline LfAPI_CS LfNTransform& Invert				(const LfNTransform&, LfNTransform&);
friend inline LfAPI_CS LfPoint&      PTrans				(const LfNTransform&, const LfPoint&,  LfPoint&);
friend inline LfAPI_CS LfPoint&      PTransNL			(const LfNTransform&, const LfPoint&,  LfPoint&);
friend inline LfAPI_CS LfVector&     VTrans				(const LfNTransform&, const LfVector&, LfVector&);
friend inline LfAPI_CS LfVector&     NTrans				(const LfNTransform&, const LfVector&, LfVector&);

friend class LfTransform;
};


//
// I M P L E M E N T A T I O N
//

#include <Lightflow/LfVectorIO.h>

#ifdef _V
#undef _V
#endif
#define _V(v, i)   (((LfVector&)v)[(i)])

#ifdef _P
#undef _P
#endif
#define _P(p, i)   (((LfPoint&)p)[(i)])

#ifdef _M
#undef _M
#endif
#define _M(m, i, j)   (((LfMatrix4&)m)[(i)][(j)])

#ifdef _M3
#undef _M3
#endif
#define _M3(m, i, j)   (((LfMatrix3&)m)[(i)][(j)])

// Temporaries
static LfFloat       __f_t1,  __f_t2,  __f_t3;
static LfVector      __v_t1,  __v_t2,  __v_t3;
static LfTransform   __t_t1,  __t_t2,  __t_t3,  __t_t4,  __t_t5,  __t_t6;
static LfNTransform  __n_t1,  __n_t2,  __n_t3,  __n_t4,  __n_t5,  __n_t6;
static LfMatrix3     __m3_t1, __m3_t2;
static LfMatrix4     __m4_t1, __m4_t2;	// NOT used !!!!!!!


//
// LfTransform METHODS
//

//
// LfTransform inline methods
//

inline LfTransform::LfTransform() { }

inline LfTransform::LfTransform(const LfTransform& a)
{
	t = a.t;
}

inline LfTransform::LfTransform(const LfNTransform& a)
{
	t = a.t;
}

inline LfTransform::LfTransform(const LfMatrix4& m)
{
	t = m;
}

inline LfTransform& LfTransform::operator = (const LfTransform& a)
{
	t = a.t;
	return *this;
}



//
// LfTransform inline functions (not members)
//

inline LfAPI_CS int operator == (const LfTransform& a, const LfTransform& b)
{
	return (a.t == b.t);
}

inline LfAPI_CS int operator != (const LfTransform& a, const LfTransform& b)
{
	return !(a == b);
}

inline LfAPI_CS LfTransform& Compose(const LfTransform& a, const LfTransform& b, LfTransform& r)
{
	// Composition makes the rightmost transformation the first applied !

	Multiply(a.t, b.t, r.t);

//	cout << "Compose() :\n";
//	cout << "a  = \n" << a.t << '\n';
//	cout << "b  = \n" << b.t << '\n';
//	cout << "ab = \n" << r.t << '\n';

	return r;
}

inline LfAPI_CS LfTransform& Invert(const LfTransform& a, LfTransform& r)
{
	Invert(a.t, r.t);
	return r;
}

inline LfAPI_CS LfPoint& PTrans(const LfTransform& a, const LfPoint& p, LfPoint& r)
{
	r[0] = _M(a.t,0,0) * _P(p,0) +                   \
	       _M(a.t,0,1) * _P(p,1) +                   \
           _M(a.t,0,2) * _P(p,2) +                   \
           _M(a.t,0,3);

	r[1] = _M(a.t,1,0) * _P(p,0) +                   \
	       _M(a.t,1,1) * _P(p,1) +                   \
           _M(a.t,1,2) * _P(p,2) +                   \
           _M(a.t,1,3);

	r[2] = _M(a.t,2,0) * _P(p,0) +                   \
	       _M(a.t,2,1) * _P(p,1) +                   \
           _M(a.t,2,2) * _P(p,2) +                   \
           _M(a.t,2,3);

	return r;
}

inline LfAPI_CS LfPoint& PTransNL(const LfTransform& a, const LfPoint& p, LfPoint& r)
{
	r[0] = _M(a.t,0,0) * _P(p,0) +                   \
	       _M(a.t,0,1) * _P(p,1) +                   \
           _M(a.t,0,2) * _P(p,2) +                   \
           _M(a.t,0,3);

	r[1] = _M(a.t,1,0) * _P(p,0) +                   \
	       _M(a.t,1,1) * _P(p,1) +                   \
           _M(a.t,1,2) * _P(p,2) +                   \
           _M(a.t,1,3);

	r[2] = _M(a.t,2,0) * _P(p,0) +                   \
	       _M(a.t,2,1) * _P(p,1) +                   \
           _M(a.t,2,2) * _P(p,2) +                   \
           _M(a.t,2,3);

//	cout << "PTransNL() :\n";
//	cout << "m   = \n" << a.t << '\n';
//	cout << "p   = \n" << p << '\n';
//	cout << "r'  = \n" << r << '\n';

	__f_t1 = 1.0/(_M(a.t,3,0) * _P(p,0) + _M(a.t,3,1) * _P(p,1) + _M(a.t,3,2) * _P(p,2) + _M(a.t,3,3));

	r[0] *= __f_t1;
	r[1] *= __f_t1;
	r[2] *= __f_t1;

//	cout << "r  = \n" << r << '\n';

	return r;
}

inline LfAPI_CS LfVector& VTrans(const LfTransform& a, const LfVector& v, LfVector& r)
{
	r[0] = _M(a.t,0,0) * _V(v,0) +                   \
	       _M(a.t,0,1) * _V(v,1) +                   \
           _M(a.t,0,2) * _V(v,2);

	r[1] = _M(a.t,1,0) * _V(v,0) +                   \
	       _M(a.t,1,1) * _V(v,1) +                   \
           _M(a.t,1,2) * _V(v,2);

	r[2] = _M(a.t,2,0) * _V(v,0) +                   \
	       _M(a.t,2,1) * _V(v,1) +                   \
           _M(a.t,2,2) * _V(v,2);

	return r;
}


//
// LfNTransform METHODS
//

//
// LfNTransform inline methods
//

inline LfNTransform::LfNTransform() { }

inline LfNTransform::LfNTransform(const LfNTransform& a)
{
	t = a.t;
	n = a.n;
}

inline LfNTransform::LfNTransform(const LfTransform& a)
{
	t = a.t;

	__m3_t1[0][0] = t[0][0];
	__m3_t1[0][1] = t[0][1];
	__m3_t1[0][2] = t[0][2];
	__m3_t1[1][0] = t[1][0];
	__m3_t1[1][1] = t[1][1];
	__m3_t1[1][2] = t[1][2];
	__m3_t1[2][0] = t[2][0];
	__m3_t1[2][1] = t[2][1];
	__m3_t1[2][2] = t[2][2];

	Invert(__m3_t1, __m3_t2);
	Transpose(__m3_t2, n);
}

inline LfNTransform::LfNTransform(const LfMatrix4& m)
{
	t = m;

	__m3_t1[0][0] = t[0][0];
	__m3_t1[0][1] = t[0][1];
	__m3_t1[0][2] = t[0][2];
	__m3_t1[1][0] = t[1][0];
	__m3_t1[1][1] = t[1][1];
	__m3_t1[1][2] = t[1][2];
	__m3_t1[2][0] = t[2][0];
	__m3_t1[2][1] = t[2][1];
	__m3_t1[2][2] = t[2][2];

	Invert(__m3_t1, __m3_t2);
	Transpose(__m3_t2, n);
}

inline LfNTransform& LfNTransform::operator = (const LfNTransform& a)
{
	t = a.t;
	n = a.n;
	return *this;
}


//
// LfNTransform inline functions (not members)
//

inline LfAPI_CS int operator == (const LfNTransform& a, const LfNTransform& b)
{
	return (a.t == b.t);
}

inline LfAPI_CS int operator != (const LfNTransform& a, const LfNTransform& b)
{
	return !(a == b);
}

inline LfAPI_CS LfNTransform& Compose(const LfNTransform& a, const LfNTransform& b, LfNTransform& r)
{
	// Composition makes the rightmost transformation the first applied !

	Multiply(a.t, b.t, r.t);

	// For normal transformation matrices, we have:
	// R  = A*B
	// N = trans(inv(R)) = trans(inv(A*B)) = trans(inv(B)*inv(A)) = trans(inv(A))*trans(inv(B))
	Multiply(a.n, b.n, r.n);

	return r;
}

inline LfAPI_CS LfNTransform& Invert(const LfNTransform& a, LfNTransform& r)
{
	Invert(a.t, r.t);
	Invert(a.n, r.n);
	return r;
}

inline LfAPI_CS LfPoint& PTrans(const LfNTransform& a, const LfPoint& p, LfPoint& r)
{
	r[0] = _M(a.t,0,0) * _P(p,0) +                   \
	       _M(a.t,0,1) * _P(p,1) +                   \
           _M(a.t,0,2) * _P(p,2) +                   \
           _M(a.t,0,3);

	r[1] = _M(a.t,1,0) * _P(p,0) +                   \
	       _M(a.t,1,1) * _P(p,1) +                   \
           _M(a.t,1,2) * _P(p,2) +                   \
           _M(a.t,1,3);

	r[2] = _M(a.t,2,0) * _P(p,0) +                   \
	       _M(a.t,2,1) * _P(p,1) +                   \
           _M(a.t,2,2) * _P(p,2) +                   \
           _M(a.t,2,3);

	return r;
}

inline LfAPI_CS LfPoint& PTransNL(const LfNTransform& a, const LfPoint& p, LfPoint& r)
{
	r[0] = _M(a.t,0,0) * _P(p,0) +                   \
	       _M(a.t,0,1) * _P(p,1) +                   \
           _M(a.t,0,2) * _P(p,2) +                   \
           _M(a.t,0,3);

	r[1] = _M(a.t,1,0) * _P(p,0) +                   \
	       _M(a.t,1,1) * _P(p,1) +                   \
           _M(a.t,1,2) * _P(p,2) +                   \
           _M(a.t,1,3);

	r[2] = _M(a.t,2,0) * _P(p,0) +                   \
	       _M(a.t,2,1) * _P(p,1) +                   \
           _M(a.t,2,2) * _P(p,2) +                   \
           _M(a.t,2,3);

	__f_t1 = 1.0/(_M(a.t,3,0) * _P(p,0) + _M(a.t,3,1) * _P(p,1) + _M(a.t,3,2) * _P(p,2) + _M(a.t,3,3));

	r[0] *= __f_t1;
	r[1] *= __f_t1;
	r[2] *= __f_t1;

	return r;
}

inline LfAPI_CS LfVector& VTrans(const LfNTransform& a, const LfVector& v, LfVector& r)
{
	r[0] = _M(a.t,0,0) * _V(v,0) +                   \
	       _M(a.t,0,1) * _V(v,1) +                   \
           _M(a.t,0,2) * _V(v,2);

	r[1] = _M(a.t,1,0) * _V(v,0) +                   \
	       _M(a.t,1,1) * _V(v,1) +                   \
           _M(a.t,1,2) * _V(v,2);

	r[2] = _M(a.t,2,0) * _V(v,0) +                   \
	       _M(a.t,2,1) * _V(v,1) +                   \
           _M(a.t,2,2) * _V(v,2);

	return r;
}

inline LfAPI_CS LfVector& NTrans(const LfNTransform& a, const LfVector& v, LfVector& r)
{
	r[0] = _M3(a.n,0,0) * _V(v,0) +                   \
	       _M3(a.n,0,1) * _V(v,1) +                   \
           _M3(a.n,0,2) * _V(v,2);

	r[1] = _M3(a.n,1,0) * _V(v,0) +                   \
	       _M3(a.n,1,1) * _V(v,1) +                   \
           _M3(a.n,1,2) * _V(v,2);

	r[2] = _M3(a.n,2,0) * _V(v,0) +                   \
	       _M3(a.n,2,1) * _V(v,1) +                   \
           _M3(a.n,2,2) * _V(v,2);

	return r;
}

inline LfAPI_CS LfVector4& LfTransform::operator [] (int i) // Added by J.Pantaleoni (14/7/97
{
	return t[i];
}

inline LfAPI_CS const LfVector4& LfTransform::Get(int i) const
{
	return t.Get(i);
}
inline LfAPI_CS const LfMatrix4& LfTransform::GetMatrix(void) const
{
	return t;
}

#undef _V
#undef _P
#undef _M
#undef _M3

#endif /* _LfTransform_h */
